/*** FKUNZIP.C - Extract files from an *uncompressed* PKZIP file.
 *   Version 1.01.19961017.
 *
 *   (C) Fedon Kadifeli, 1996.
 *
 *   Usage:     FKUNZIP filename
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>


/***
 * constants
 */

#define BUFLEN (10*1024)

unsigned char const cmagic[] = {  'P', 'K', 0x03, 0x04 };

char const * extlist[] = {
 ".html", ".htm", ".txt",
 ".pl", ".sh",
 ".c", ".cpp",
 ".h", ".hpp"
};


/***
 * types
 */

struct head {
  unsigned char magic [4];
  unsigned char unk   [6];
  unsigned char dt    [4];
  unsigned char csum  [4];
  unsigned char csize [4];
  unsigned char osize [4];
  unsigned char nlen  [4];
};
typedef unsigned long u_long;


/***
 * functions
 */

void dumphex (unsigned char *cp, int i)
{
  while (i-->0)
    printf ("%02x ", *cp++);
  printf ("\n");
} /* dumphex */


void chfiledt (char *fn, time_t mdt)
{
  struct utimbuf utb;

  time(&utb.actime);
  utb.modtime = mdt;
  utime(fn, &utb);
}  /* chfiledt */


u_long getlong (unsigned char s[4])
{
  u_long l = 0;
  int c;

  for (c=3; c>=0; c--) {
    l <<= 8;
    l |= s[c];
  }
  return l;
} /* getlong */


time_t gettime (unsigned char dt[4])
{
  struct tm  stm = { 0 };
  u_long l = getlong(dt);

  stm.tm_sec = (l&0x1f) * 2;
  stm.tm_min = (l>>5) & 0x3f;
  stm.tm_hour = (l>>11) & 0x1f;
  stm.tm_mday = (l>>16) & 0x1f;
  stm.tm_mon =  ((l>>21) & 0xf) - 1;
  stm.tm_year = ((l>>25) & 0x7f) + 80;
  return mktime (&stm);
} /* gettime */


u_long delret (register char * src, u_long len)
{
  register char * dst;
  register u_long i;

  for (i=0, dst=src; i<len; i++)
    if (*src == '\r')
      src++;
    else
      *dst++ = *src++;
  return len+dst-src;
} /* delret */


int istext (char const *fn)
{
  int i;
  int l = strlen(fn);

  for (i=0; i<sizeof(extlist)/sizeof(*extlist); i++)
    if (strcmp(fn+l-strlen(extlist[i]), extlist[i])==0)
      return 1;
  return 0;
} /* istext */


int procfile (FILE *inf, char *fn, u_long fsize, time_t fdt, int fskip)
{
  u_long cc, cr;
  FILE *outf;
  static char buf [BUFLEN];
  int txtfile;

  if (!fskip) {
    outf = fopen (fn, "wb");
    if (outf == NULL) {
      printf ("\n *** Skipping file: %s : %s", fn, strerror(errno));
      fskip = 1;
    }
  }
  if (!fskip) {
    txtfile = istext (fn);
    printf ("(%s)", txtfile ? "text" : "binary");
  }
  for (cc=0; cc<fsize; cc+=cr) {
    cr = fsize - cc;
    if (cr > BUFLEN)
      cr = BUFLEN;
    if (fread (buf, 1, (size_t)cr, inf) != cr) {
      fprintf (stderr, "\n *** Unable to read data for file %s\n", fn);
      exit (5);
    }
    if (!fskip)
      fwrite (buf, 1, (size_t)((txtfile) ? delret (buf, cr) : cr), outf);
  } /* for */
  if (!fskip) {
    fclose (outf);
    chfiledt (fn, fdt);
  }
  return fskip;
} /* procfile */


void main (int argc, char ** argv)
{
  struct head   h;
  FILE *        inf;
  time_t        dt;
  u_long        csum, csize, osize, nlen;
  char          filename [FILENAME_MAX+1], tbuf[40], * cp;
  int           skip, count=0, scount=0, dcount=0;

  if (argc != 2) {
    fprintf (stderr, "Usage: %s filename\n", argv[0]);
    exit (1);
  }
  if ((inf = fopen (argv[1], "rb")) == NULL) {
    perror (argv[1]);
    exit (2);
  }
  while ((fread ((void *)&h, 1, sizeof (h), inf) == sizeof(h)) &&
         (memcmp (h.magic, cmagic, sizeof (cmagic)) == 0)) {
    dt = gettime (h.dt);
    csum = getlong (h.csum);
    csize = getlong (h.csize);
    osize = getlong (h.osize);
    nlen = getlong (h.nlen);
    if (nlen > FILENAME_MAX) {
      fprintf (stderr, "\n *** File name too long\n");
      exit (3);
    }
    if (fread (filename, 1, (size_t)nlen, inf) != nlen) {
      fprintf (stderr, "\n *** Unable to read file name\n");
      exit (4);
    }
    filename [nlen] = '\0';
    for (cp=filename; *cp; cp++)
      *cp = tolower (*cp);
    strftime (tbuf, sizeof(tbuf)-1,
      "%Y/%m/%d %H:%M:%S", localtime(&dt));
    printf ("%s %08lx %8ld %8ld %s ", tbuf, csum, csize, osize, filename);
    if (filename[nlen-1]=='/') {        /* if directory */
      printf ("(dir)");
      dcount++;
      filename [nlen-1] = '\0';
      mkdir (filename, 0777);
      chfiledt (filename, dt);
    } else { /* if file */
      skip = 0;
      count++;
      if (csize != osize) {
        printf ("\n *** Skipping compressed file: %s", filename);
        skip = 1;
      }
      scount += procfile (inf, filename, csize, dt, skip);
    }
    printf ("\n");
  } /* while */
  fclose (inf);
  if (count == 0)
    fprintf (stderr, "\n *** No files found to extract\n");
  else 
    printf ("\nTotal %d files: %d files skipped, %d files processed.\n",
      count, scount, count-scount);
  if (dcount != 0)
    printf ("Total %d directories.\n", dcount);
} /* main */
